"""GasMaterial module."""

import collections

import numpy as np
from sigfig import round
from validator_collection import validators

from .material_base import MaterialBase


class GasMaterial(MaterialBase):
    """Gas Materials.

    .. image:: ../images/template/materials-gas.png
    """
    _CREATED_OBJECTS = []

    __slots__ = ("_type", "_conductivity", "_density")

    _GASTYPES = ("air", "argon", "krypton", "xenon", "sf6")

    def __init__(
        self, Name, Conductivity=None, Density=None, Category="Gases", **kwargs
    ):
        """Initialize object with parameters.

        Args:
            Name (str): The name of the GasMaterial.
            Conductivity (float): Thermal conductivity (W/m-K).
            Density (float): A number representing the density of the material
                in kg/m3. This is essentially the mass of one cubic meter of the
                material.
            Category (str): Category is set as "Gases" for GasMaterial.
            **kwargs: keywords passed to the MaterialBase constructor.
        """
        self.Name = Name
        super(GasMaterial, self).__init__(self.Name, Category=Category, **kwargs)
        self.Type = Name.upper()
        self.Conductivity = Conductivity
        self.Density = Density

        # Only at the end append self to _CREATED_OBJECTS
        self._CREATED_OBJECTS.append(self)

    @property
    def Name(self):
        """Get or set the name of the GasMaterial.

        Choices are ("Air", "Argon", "Krypton", "Xenon").
        """
        return self._name

    @Name.setter
    def Name(self, value):
        assert value.lower() in self._GASTYPES, (
            f"Invalid value '{value}' for material gas type. Gas type must be one "
            f"of the following:\n{self._GASTYPES}"
        )
        self._type = value.upper()
        self._name = value

    @property
    def Type(self):
        """Get or set the gas type. Alias of Name.

        Choices are ("Air", "Argon", "Krypton", "Xenon").
        """
        return self._type

    @Type.setter
    def Type(self, value):
        assert value.lower() in self._GASTYPES, (
            f"Invalid value '{value}' for material gas type. Gas type must be one "
            f"of the following:\n{self._GASTYPES}"
        )
        self._type = value.upper()
        self._name = value

    @property
    def Conductivity(self):
        """Get or set the conductivity of the gas at 0C [W/m-K]."""
        return self._conductivity

    @Conductivity.setter
    def Conductivity(self, value):
        if value is not None:
            self._conductivity = validators.float(value, minimum=0)
        else:
            self._conductivity = self.conductivity_at_temperature(273.15)

    @property
    def Density(self):
        """Get or set the density of the gas."""
        return self._density

    @Density.setter
    def Density(self, value):
        """Density of the gas at 0C and sea-level pressure [kg/m3]."""
        if value is not None:
            self._density = validators.float(value, minimum=0)
        else:
            self._density = self.density_at_temperature(273.15)

    @property
    def molecular_weight(self):
        """Get the molecular weight [kg/mol]."""
        import CoolProp.CoolProp as CP

        return CP.PropsSI("molemass", self.Type)

    @property
    def specific_heat(self):
        """Get the material layer's specific heat at 0C [J/kg-K]."""
        return self.specific_heat_at_temperature(273.15)

    def duplicate(self):
        """Get copy of self."""
        return self.__copy__()

    @classmethod
    def from_dict(cls, data, **kwargs):
        """Create a GasMaterial from a dictionary.

        Args:
            data (dict): A python dictionary.
            **kwargs: keywords passed the MaterialBase constructor.
        """
        _id = data.pop("$id")
        return cls(id=_id, **data, **kwargs)

    def to_dict(self):
        """Return GasMaterial dictionary representation."""
        self.validate()  # Validate object before trying to get json format

        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["Category"] = self.Category
        data_dict["Type"] = self.Type
        data_dict["Conductivity"] = round(self.Conductivity, sigfigs=2)
        data_dict["Cost"] = self.Cost
        data_dict["Density"] = self.Density
        data_dict["EmbodiedCarbon"] = self.EmbodiedCarbon
        data_dict["EmbodiedEnergy"] = self.EmbodiedEnergy
        data_dict["SubstitutionRatePattern"] = self.SubstitutionRatePattern
        data_dict["SubstitutionTimestep"] = self.SubstitutionTimestep
        data_dict["TransportCarbon"] = self.TransportCarbon
        data_dict["TransportDistance"] = self.TransportDistance
        data_dict["TransportEnergy"] = self.TransportEnergy
        data_dict["Comments"] = validators.string(self.Comments, allow_empty=True)
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = self.Name

        return data_dict

    def to_epbunch(self, idf, thickness):
        """Convert self to an epbunch given an idf model and a thickness.

        Args:
            idf (IDF): An IDF model.
            thickness (float): the thickness of the material.

        .. code-block:: python

            WindowMaterial:Gas,
                AIR_0.006_B_Dbl_Air_Cl,    !- Name
                AIR,                      !- Gas Type
                0.006;                    !- Thickness

        Returns:
            EpBunch: The EpBunch object added to the idf model.
        """
        return idf.newidfobject(
            "WINDOWMATERIAL:GAS",
            Name=self.Name,
            Gas_Type=self.Type,
            Thickness=thickness,
        )

    def mapping(self, validate=False):
        """Get a dict based on the object properties, useful for dict repr.

        Args:
            validate (bool): If True, try to validate object before returning the
                mapping.
        """
        if validate:
            self.validate()

        return dict(
            Category=self.Category,
            Type=self.Type,
            Conductivity=self.Conductivity,
            Cost=self.Cost,
            Density=self.Density,
            EmbodiedCarbon=self.EmbodiedCarbon,
            EmbodiedEnergy=self.EmbodiedEnergy,
            SubstitutionRatePattern=self.SubstitutionRatePattern,
            SubstitutionTimestep=self.SubstitutionTimestep,
            TransportCarbon=self.TransportCarbon,
            TransportDistance=self.TransportDistance,
            TransportEnergy=self.TransportEnergy,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    def density_at_temperature(self, t_kelvin, pressure=101325):
        """Get the density of the gas [kg/m3] at a given temperature and pressure.

        This method uses CoolProp to get the density.

        Args:
            t_kelvin (float): The average temperature of the gas cavity in Kelvin.
            pressure (float): The average pressure of the gas cavity in Pa.
                Default is 101325 Pa for standard pressure at sea level.
        """
        import CoolProp.CoolProp as CP

        return CP.PropsSI("Dmass", "T", t_kelvin, "P", pressure, self.Type)

    def specific_heat_at_temperature(self, t_kelvin, pressure=101325):
        """Get the specific heat of the gas [J/(kg-K)] at a given Kelvin temperature.

        This method uses CoolProp to get the density.

        Args:
            t_kelvin (float): The average temperature of the gas cavity in Kelvin.
            pressure (float): The average pressure of the gas cavity in Pa.
                Default is 101325 Pa for standard pressure at sea level.
        """
        import CoolProp.CoolProp as CP

        return CP.PropsSI("Cpmass", "T", t_kelvin, "P", pressure, self.Type)

    def viscosity_at_temperature(self, t_kelvin, pressure=101325):
        """Get the viscosity of the gas [kg/m-s] at a given Kelvin temperature.

        This method uses CoolProp to get the density.

        Args:
            t_kelvin (float): The average temperature of the gas cavity in Kelvin.
            pressure (float): The average pressure of the gas cavity in Pa.
                Default is 101325 Pa for standard pressure at sea level.
        """
        import CoolProp.CoolProp as CP

        try:
            return CP.PropsSI("viscosity", "T", t_kelvin, "P", pressure, self.Type)
        except ValueError:
            # ValueError: Viscosity model is not available for Krypton, Xenon
            return {"krypton": 2.3219e-5, "xenon": 2.1216e-5}[self.Type.lower()]

    def conductivity_at_temperature(self, t_kelvin, pressure=101325):
        """Get the conductivity of the gas [W/(m-K)] at a given Kelvin temperature.

        This method uses CoolProp to get the density. Note that the thermal
        conductivity model is not available for Krypton, Xenon gases. Values from the
        literature are used instead.

        Args:
            t_kelvin (float): The average temperature of the gas cavity in Kelvin.
            pressure (float): The average pressure of the gas cavity in Pa.
                Default is 101325 Pa for standard pressure at sea level.
        """
        import CoolProp.CoolProp as CP

        try:
            return CP.PropsSI("conductivity", "T", t_kelvin, "P", pressure, self.Type)
        except ValueError:
            # ValueError: Thermal conductivity model is not available for Krypton, Xenon
            return {"krypton": 0.00943, "xenon": 5.65e-3}[self.Type.lower()]

    def __hash__(self):
        """Return the hash value of self."""
        return hash(self.id)

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, GasMaterial):
            return NotImplemented
        else:
            return all(
                [
                    self.Category == other.Category,
                    self.Type == other.Type,
                    self.Conductivity == other.Conductivity,
                    self.Cost == other.Cost,
                    self.Density == other.Density,
                    self.EmbodiedCarbon == other.EmbodiedCarbon,
                    self.EmbodiedEnergy == other.EmbodiedEnergy,
                    np.array_equal(
                        self.SubstitutionRatePattern, other.SubstitutionRatePattern
                    ),
                    self.SubstitutionTimestep == other.SubstitutionTimestep,
                    self.TransportCarbon == other.TransportCarbon,
                    self.TransportDistance == other.TransportDistance,
                    self.TransportEnergy == other.TransportEnergy,
                ]
            )

    def __copy__(self):
        """Create a copy of self."""
        return self.__class__(**self.mapping(validate=False))



"""archetypal WindowSettings."""

import collections
import logging as lg
from copy import copy
from functools import reduce

from validator_collection import checkers, validators
from validator_collection.errors import EmptyValueError

from archetypal.template.constructions.window_construction import (
    ShadingType,
    WindowConstruction,
    WindowType,
)
from archetypal.template.schedule import UmiSchedule
from archetypal.template.umi_base import UmiBase
from archetypal.utils import log, timeit


class WindowSetting(UmiBase):
    """Defines the various window-related properties of a :class:`Zone`.

    Control natural ventilation, shading and airflow networks and more using this
    class. This class serves the same role as the ZoneInformation>Windows tab in the
    UMI TemplateEditor.

    .. image:: ../images/template/zoneinfo-windows.png

    Hint:
        The WindowSetting class implements two constructors that are tailored to
        the eppy_ scripting language:

        - :func:`from_construction` and
        - :func:`from_surface`.

    .. _eppy : https://eppy.readthedocs.io/en/latest/
    """
    _CREATED_OBJECTS = []

    __slots__ = (
        "_operable_area",
        "_afn_discharge_c",
        "_afn_temp_setpoint",
        "_shading_system_setpoint",
        "_shading_system_transmittance",
        "_zone_mixing_availability_schedule",
        "_shading_system_availability_schedule",
        "_construction",
        "_afn_window_availability",
        "_is_shading_system_on",
        "_is_virtual_partition",
        "_is_zone_mixing_on",
        "_shading_system_type",
        "_type",
        "_zone_mixing_delta_temperature",
        "_zone_mixing_flow_rate",
        "_area",
    )

    def __init__(
        self,
        Name,
        Construction=None,
        OperableArea=0.8,
        AfnWindowAvailability=None,
        AfnDischargeC=0.65,
        AfnTempSetpoint=20,
        IsVirtualPartition=False,
        IsShadingSystemOn=False,
        ShadingSystemAvailabilitySchedule=None,
        ShadingSystemSetpoint=180,
        ShadingSystemTransmittance=0.5,
        ShadingSystemType=ShadingType.ExteriorShade,
        Type=WindowType.External,
        IsZoneMixingOn=False,
        ZoneMixingAvailabilitySchedule=None,
        ZoneMixingDeltaTemperature=2,
        ZoneMixingFlowRate=0.001,
        area=1,
        **kwargs,
    ):
        """Initialize a WindowSetting using default values.

        Args:
            Construction (WindowConstruction): The window construction.
            OperableArea (float): The operable window area as a ratio of total
                window area. eg. 0.8 := 80% of the windows area is operable.
            AfnWindowAvailability (UmiSchedule): The Airflow Network availability
                schedule.
            AfnDischargeC (float): Airflow Network Discharge Coefficient.
                Default = 0.65.
            AfnTempSetpoint (float): Airflow Network Temperature Setpoint.
                Default = 20 degreeC.
            IsVirtualPartition (bool): Virtual Partition.
            IsShadingSystemOn (bool): Shading is used. Default is False.
            ShadingSystemAvailabilitySchedule (UmiSchedule): Shading system
                availability schedule.
            ShadingSystemSetpoint (float): Shading system setpoint in units of
                W/m2. Default = 180 W/m2.
            ShadingSystemTransmittance (float): Shading system transmittance.
                Default = 0.5.
            ShadingSystemType (int): Shading System Type. 0 = ExteriorShade, 1 =
                InteriorShade.
            Type (int):
            IsZoneMixingOn (bool): Zone mixing.
            ZoneMixingAvailabilitySchedule (UmiSchedule): Zone mixing
                availability schedule.
            ZoneMixingDeltaTemperature (float): Zone mixing delta
            ZoneMixingFlowRate (float): Zone mixing flow rate in units of m3/m2.
                Default = 0.001 m3/m2.
            **kwargs: other keywords passed to the constructor.
        """
        super(WindowSetting, self).__init__(Name, **kwargs)

        self.ShadingSystemAvailabilitySchedule = ShadingSystemAvailabilitySchedule
        self.Construction = Construction
        self.AfnWindowAvailability = AfnWindowAvailability
        self.AfnDischargeC = AfnDischargeC
        self.AfnTempSetpoint = AfnTempSetpoint
        self.IsShadingSystemOn = IsShadingSystemOn
        self.IsVirtualPartition = IsVirtualPartition
        self.IsZoneMixingOn = IsZoneMixingOn
        self.OperableArea = OperableArea
        self.ShadingSystemSetpoint = ShadingSystemSetpoint
        self.ShadingSystemTransmittance = ShadingSystemTransmittance
        self.ShadingSystemType = ShadingSystemType
        self.Type = Type
        self.ZoneMixingDeltaTemperature = ZoneMixingDeltaTemperature
        self.ZoneMixingFlowRate = ZoneMixingFlowRate
        self.ZoneMixingAvailabilitySchedule = ZoneMixingAvailabilitySchedule

        self.area = area

        # Only at the end append self to _CREATED_OBJECTS
        self._CREATED_OBJECTS.append(self)

    @property
    def area(self):
        """Get or set the area of the zone associated to this object [m²]."""
        return self._area

    @area.setter
    def area(self, value):
        self._area = validators.float(value, minimum=0)

    @property
    def OperableArea(self):
        """Get or set the operable area ratio [-]."""
        return self._operable_area

    @OperableArea.setter
    def OperableArea(self, value):
        self._operable_area = validators.float(value, minimum=0, maximum=1)

    @property
    def AfnDischargeC(self):
        """Get or set the air flow network discarge coefficient."""
        return self._afn_discharge_c

    @AfnDischargeC.setter
    def AfnDischargeC(self, value):
        self._afn_discharge_c = validators.float(value, minimum=0, maximum=1)

    @property
    def AfnTempSetpoint(self):
        """Get or set the air flow network setpoint temperature [degC]."""
        return self._afn_temp_setpoint

    @AfnTempSetpoint.setter
    def AfnTempSetpoint(self, value):
        self._afn_temp_setpoint = validators.float(value, minimum=-100, maximum=100)

    @property
    def AfnWindowAvailability(self):
        """Get or set the air flow network window availability schedule."""
        return self._afn_window_availability

    @AfnWindowAvailability.setter
    def AfnWindowAvailability(self, value):
        if value is not None:
            assert isinstance(value, UmiSchedule), (
                f"Input error with value {value}. AfnWindowAvailability must "
                f"be an UmiSchedule, not a {type(value)}"
            )
        self._afn_window_availability = value

    @property
    def ShadingSystemType(self):
        """Get or set the shading system type [enum]."""
        return self._shading_system_type

    @ShadingSystemType.setter
    def ShadingSystemType(self, value):
        if checkers.is_string(value):
            assert ShadingType[value], (
                f"Input value error for '{value}'. "
                f"Expected one of {tuple(a for a in ShadingType)}"
            )
            self._shading_system_type = ShadingType[value]
        elif checkers.is_numeric(value):
            assert ShadingType(value), (
                f"Input value error for '{value}'. "
                f"Expected one of {tuple(a for a in ShadingType)}"
            )
            self._shading_system_type = ShadingType(value)
        elif isinstance(value, ShadingType):
            self._shading_system_type = value

    @property
    def ShadingSystemSetpoint(self):
        """Get or set the shading system setpoint [W/m2]."""
        return self._shading_system_setpoint

    @ShadingSystemSetpoint.setter
    def ShadingSystemSetpoint(self, value):
        self._shading_system_setpoint = validators.float(value, minimum=0)

    @property
    def ShadingSystemTransmittance(self):
        """Get or set the shading system transmittance [-]."""
        return self._shading_system_transmittance

    @ShadingSystemTransmittance.setter
    def ShadingSystemTransmittance(self, value):
        self._shading_system_transmittance = validators.float(
            value, minimum=0, maximum=1
        )

    @property
    def ShadingSystemAvailabilitySchedule(self):
        """Get or set the shading system availability schedule."""
        return self._shading_system_availability_schedule

    @ShadingSystemAvailabilitySchedule.setter
    def ShadingSystemAvailabilitySchedule(self, value):
        if value is not None:
            assert isinstance(value, UmiSchedule), (
                f"Input error with value {value}. ZoneMixingAvailabilitySchedule must "
                f"be an UmiSchedule, not a {type(value)}"
            )
        self._shading_system_availability_schedule = value

    @property
    def IsShadingSystemOn(self):
        """Get or set the use of the shading system."""
        return self._is_shading_system_on

    @IsShadingSystemOn.setter
    def IsShadingSystemOn(self, value):
        assert isinstance(value, bool), (
            f"Input error with value {value}. IsShadingSystemOn must "
            f"be a boolean, not a {type(value)}"
        )
        self._is_shading_system_on = value

    @property
    def ZoneMixingAvailabilitySchedule(self):
        """Get or set the zone mixing availability schedule."""
        return self._zone_mixing_availability_schedule

    @ZoneMixingAvailabilitySchedule.setter
    def ZoneMixingAvailabilitySchedule(self, value):
        if value is not None:
            assert isinstance(value, UmiSchedule), (
                f"Input error with value {value}. ZoneMixingAvailabilitySchedule must "
                f"be an UmiSchedule, not a {type(value)}"
            )
        self._zone_mixing_availability_schedule = value

    @property
    def ZoneMixingDeltaTemperature(self):
        """Get or set the zone mixing delta temperature."""
        return self._zone_mixing_delta_temperature

    @ZoneMixingDeltaTemperature.setter
    def ZoneMixingDeltaTemperature(self, value):
        self._zone_mixing_delta_temperature = validators.float(value, minimum=0)

    @property
    def Construction(self):
        """Get or set the window construction."""
        return self._construction

    @Construction.setter
    def Construction(self, value):
        if value is not None:
            assert isinstance(value, WindowConstruction), (
                f"Input error with value {value}. Construction must "
                f"be an WindowConstruction, not a {type(value)}"
            )
        self._construction = value

    @property
    def IsVirtualPartition(self):
        """Get or set the state of the virtual partition."""
        return self._is_virtual_partition

    @IsVirtualPartition.setter
    def IsVirtualPartition(self, value):
        assert isinstance(value, bool), (
            f"Input error with value {value}. IsVirtualPartition must "
            f"be a boolean, not a {type(value)}"
        )
        self._is_virtual_partition = value

    @property
    def IsZoneMixingOn(self):
        """Get or set mixing in zone."""
        return self._is_zone_mixing_on

    @IsZoneMixingOn.setter
    def IsZoneMixingOn(self, value):
        assert isinstance(value, bool), (
            f"Input error with value {value}. IsZoneMixingOn must "
            f"be a boolean, not a {type(value)}"
        )
        self._is_zone_mixing_on = value

    @property
    def ZoneMixingFlowRate(self):
        """Get or set the zone mixing flow rate [m3/s]."""
        return self._zone_mixing_flow_rate

    @ZoneMixingFlowRate.setter
    def ZoneMixingFlowRate(self, value):
        self._zone_mixing_flow_rate = validators.float(value, minimum=0)

    @property
    def Type(self):
        """Get or set the window type [enum]."""
        return self._type

    @Type.setter
    def Type(self, value):
        if checkers.is_string(value):
            assert WindowType[value], (
                f"Input value error for '{value}'. "
                f"Expected one of {tuple(a for a in WindowType)}"
            )
            self._type = WindowType[value]
        elif checkers.is_numeric(value):
            assert WindowType(value), (
                f"Input value error for '{value}'. "
                f"Expected one of {tuple(a for a in WindowType)}"
            )
            self._type = WindowType(value)
        elif isinstance(value, WindowType):
            self._type = value

    def __add__(self, other):
        """Combine self and other."""
        return self.combine(other)

    def __repr__(self):
        """Return a representation of self."""
        return super(WindowSetting, self).__repr__()

    def __str__(self):
        """Return string representation."""
        return repr(self)

    def __hash__(self):
        """Return the hash value of self."""
        return hash(self.id)

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, WindowSetting):
            return NotImplemented
        else:
            return all(
                [
                    self.Construction == other.Construction,
                    self.OperableArea == other.OperableArea,
                    self.AfnWindowAvailability == other.AfnWindowAvailability,
                    self.AfnDischargeC == other.AfnDischargeC,
                    self.AfnTempSetpoint == other.AfnTempSetpoint,
                    self.IsVirtualPartition == other.IsVirtualPartition,
                    self.IsShadingSystemOn == other.IsShadingSystemOn,
                    self.ShadingSystemAvailabilitySchedule
                    == other.ShadingSystemAvailabilitySchedule,
                    self.ShadingSystemSetpoint == other.ShadingSystemSetpoint,
                    self.ShadingSystemTransmittance == other.ShadingSystemTransmittance,
                    self.ShadingSystemType == other.ShadingSystemType,
                    self.Type == other.Type,
                    self.IsZoneMixingOn == other.IsZoneMixingOn,
                    self.ZoneMixingAvailabilitySchedule
                    == other.ZoneMixingAvailabilitySchedule,
                    self.ZoneMixingDeltaTemperature == other.ZoneMixingDeltaTemperature,
                    self.ZoneMixingFlowRate == other.ZoneMixingFlowRate,
                ]
            )

    @classmethod
    def generic(cls, Name):
        """Initialize a generic window with SHGC=0.704, UFactor=2.703, Tvis=0.786.

        Args:
            Name (str): Name of the WindowSetting
        """
        construction = WindowConstruction.from_shgc(
            "SimpleWindow:SINGLE PANE HW WINDOW",
            u_factor=2.703,
            solar_heat_gain_coefficient=0.704,
            visible_transmittance=0.786,
        )
        return WindowSetting(Name, Construction=construction)

    @classmethod
    def from_construction(cls, Construction, **kwargs):
        """Make a :class:`WindowSetting` directly from a Construction_ object.

        .. _Construction : https://bigladdersoftware.com/epx/docs/8-9/input-output-reference/group-surface-construction-elements.html#construction-000

        Examples:
            >>> from archetypal import IDF
            >>> from archetypal.template.window_setting import WindowSetting
            >>> # Given an IDF object
            >>> idf = IDF("idfname.idf")
            >>> constr = idf.getobject('CONSTRUCTION',
            >>>                              'AEDG-SmOffice 1A Window Fixed')
            >>> WindowSetting.from_construction(
            >>>     Name='test_window',
            >>>     Construction=constr
            >>> )

        Args:
            Construction (EpBunch): The construction name for this window.
            **kwargs: Other keywords passed to the constructor.

        Returns:
            (windowSetting): The window setting object.
        """
        name = kwargs.pop("Name", Construction.Name + "_Window")
        construction = WindowConstruction.from_epbunch(Construction, **kwargs)
        AfnWindowAvailability = UmiSchedule.constant_schedule()
        ShadingSystemAvailabilitySchedule = UmiSchedule.constant_schedule()
        ZoneMixingAvailabilitySchedule = UmiSchedule.constant_schedule()
        return cls(
            Name=name,
            Construction=construction,
            AfnWindowAvailability=AfnWindowAvailability,
            ShadingSystemAvailabilitySchedule=ShadingSystemAvailabilitySchedule,
            ZoneMixingAvailabilitySchedule=ZoneMixingAvailabilitySchedule,
            **kwargs,
        )

    @classmethod
    def from_surface(cls, surface, **kwargs):
        """Build a WindowSetting object from a FenestrationSurface:Detailed_.

        This constructor will detect common window constructions and
        shading devices. Supported Shading and Natural Air flow EnergyPlus
        objects are: WindowProperty:ShadingControl_,
        AirflowNetwork:MultiZone:Surface_.

        Important:
            If an EnergyPlus object is not supported, eg.:
            AirflowNetwork:MultiZone:Component:DetailedOpening_, only a warning
            will be issued in the console for the related object instance and
            default values will be automatically used.

        .. _FenestrationSurface:Detailed:
           https://bigladdersoftware.com/epx/docs/8-9/input-output-reference
           /group-thermal-zone-description-geometry.html
           #fenestrationsurfacedetailed
        .. _WindowProperty:ShadingControl:
           https://bigladdersoftware.com/epx/docs/8-9/input-output-reference
           /group-thermal-zone-description-geometry.html
           #windowpropertyshadingcontrol
        .. _AirflowNetwork:MultiZone:Surface:
           https://bigladdersoftware.com/epx/docs/8-9/input-output-reference
           /group-airflow-network.html#airflownetworkmultizonesurface
        .. _AirflowNetwork:MultiZone:Component:DetailedOpening:
           https://bigladdersoftware.com/epx/docs/8-9/input-output-reference
           /group-airflow-network.html
           #airflownetworkmultizonecomponentdetailedopening

        Args:
            surface (EpBunch): The FenestrationSurface:Detailed_ object.

        Returns:
            (WindowSetting): The window setting object.
        """
        if surface.key.upper() == "FENESTRATIONSURFACE:DETAILED":
            if not surface.Surface_Type.lower() == "window":
                return  # Other surface types (Doors, GlassDoors, etc.) are ignored.
            construction = surface.get_referenced_object("Construction_Name")
            if construction is None:
                construction = surface.theidf.getobject(
                    "CONSTRUCTION", surface.Construction_Name
                )
            construction = WindowConstruction.from_epbunch(construction)
            shading_control = surface.get_referenced_object("Shading_Control_Name")
        elif surface.key.upper() == "WINDOW":
            construction = surface.get_referenced_object("Construction_Name")
            construction = WindowConstruction.from_epbunch(construction)
            shading_control = next(
                iter(
                    surface.getreferingobjs(
                        iddgroups=["Thermal Zones and Surfaces"],
                        fields=[f"Fenestration_Surface_{i}_Name" for i in range(1, 10)],
                    )
                ),
                None,
            )
        elif surface.key.upper() == "DOOR":
            return  # Simply skip doors.
        else:
            raise ValueError(
                f"A window of type {surface.key} is not yet supported. "
                f"Please contact developers"
            )

        attr = {}
        if shading_control:
            # a WindowProperty:ShadingControl_ object can be attached to
            # this window
            attr["IsShadingSystemOn"] = True
            if shading_control["Setpoint"] != "":
                attr["ShadingSystemSetpoint"] = shading_control["Setpoint"]
            shade_mat = shading_control.get_referenced_object(
                "Shading_Device_Material_Name"
            )
            # get shading transmittance
            if shade_mat:
                attr["ShadingSystemTransmittance"] = shade_mat["Visible_Transmittance"]
            # get shading control schedule
            if shading_control["Shading_Control_Is_Scheduled"].upper() == "YES":
                name = shading_control["Schedule_Name"]
                attr["ShadingSystemAvailabilitySchedule"] = UmiSchedule.from_epbunch(
                    surface.theidf.schedules_dict[name.upper()]
                )
            else:
                # Determine which behavior of control
                shade_ctrl_type = shading_control["Shading_Control_Type"]
                if shade_ctrl_type.lower() == "alwaysoff":
                    attr[
                        "ShadingSystemAvailabilitySchedule"
                    ] = UmiSchedule.constant_schedule(value=0, name="AlwaysOff")
                elif shade_ctrl_type.lower() == "alwayson":
                    attr[
                        "ShadingSystemAvailabilitySchedule"
                    ] = UmiSchedule.constant_schedule()
                else:
                    log(
                        'Window "{}" uses a  window control type that '
                        'is not supported: "{}". Reverting to '
                        '"AlwaysOn"'.format(surface.Name, shade_ctrl_type),
                        lg.WARN,
                    )
                    attr[
                        "ShadingSystemAvailabilitySchedule"
                    ] = UmiSchedule.constant_schedule()
            # get shading type
            if shading_control["Shading_Type"] != "":
                mapping = {
                    "InteriorShade": ShadingType(1),
                    "ExteriorShade": ShadingType(0),
                    "ExteriorScreen": ShadingType(0),
                    "InteriorBlind": ShadingType(1),
                    "ExteriorBlind": ShadingType(0),
                    "BetweenGlassShade": ShadingType(0),
                    "BetweenGlassBlind": ShadingType(0),
                    "SwitchableGlazing": ShadingType(0),
                }
                attr["ShadingSystemType"] = mapping[shading_control["Shading_Type"]]
        else:
            # Set default schedules
            attr["ShadingSystemAvailabilitySchedule"] = UmiSchedule.constant_schedule()

        # get airflow network
        afn = next(
            iter(
                surface.getreferingobjs(
                    iddgroups=["Natural Ventilation and Duct Leakage"],
                    fields=["Surface_Name"],
                )
            ),
            None,
        )

        if afn:
            attr["OperableArea"] = afn.WindowDoor_Opening_Factor_or_Crack_Factor
            leak = afn.get_referenced_object("Leakage_Component_Name")
            name = afn["Venting_Availability_Schedule_Name"]
            if name != "":
                attr["AfnWindowAvailability"] = UmiSchedule.from_epbunch(
                    surface.theidf.schedules_dict[name.upper()]
                )
            else:
                attr["AfnWindowAvailability"] = UmiSchedule.constant_schedule()
            name = afn["Ventilation_Control_Zone_Temperature_Setpoint_Schedule_Name"]
            if name != "":
                attr["AfnTempSetpoint"] = UmiSchedule(
                    Name=name, idf=surface.theidf
                ).mean
            else:
                pass  # uses default

            if (
                leak.key.upper()
                == "AIRFLOWNETWORK:MULTIZONE:SURFACE:EFFECTIVELEAKAGEAREA"
            ):
                attr["AfnDischargeC"] = leak["Discharge_Coefficient"]
            elif (
                leak.key.upper()
                == "AIRFLOWNETWORK:MULTIZONE:COMPONENT:HORIZONTALOPENING"
            ):
                log(
                    '"{}" is not fully supported. Reverting to '
                    'defaults for object "{}"'.format(leak.key, cls.mro()[0].__name__),
                    lg.WARNING,
                )
            elif leak.key.upper() == "AIRFLOWNETWORK:MULTIZONE:SURFACE:CRACK":
                log(
                    '"{}" is not fully supported. Rerverting to '
                    'defaults for object "{}"'.format(leak.key, cls.mro()[0].__name__),
                    lg.WARNING,
                )
            elif (
                leak.key.upper() == "AIRFLOWNETWORK:MULTIZONE:COMPONENT:DETAILEDOPENING"
            ):
                log(
                    '"{}" is not fully supported. Rerverting to '
                    'defaults for object "{}"'.format(leak.key, cls.mro()[0].__name__),
                    lg.WARNING,
                )
            elif (
                leak.key.upper() == "AIRFLOWNETWORK:MULTIZONE:COMPONENT:ZONEEXHAUSTFAN"
            ):
                log(
                    '"{}" is not fully supported. Rerverting to '
                    'defaults for object "{}"'.format(leak.key, cls.mro()[0].__name__),
                    lg.WARNING,
                )
            elif leak.key.upper() == "AIRFLOWNETWORK:MULTIZONE:COMPONENT:SIMPLEOPENING":
                log(
                    '"{}" is not fully supported. Rerverting to '
                    'defaults for object "{}"'.format(leak.key, cls.mro()[0].__name__),
                    lg.WARNING,
                )
        else:
            attr["AfnWindowAvailability"] = UmiSchedule.constant_schedule(
                value=0, Name="AlwaysOff"
            )
        # Todo: Zone Mixing is always off
        attr["ZoneMixingAvailabilitySchedule"] = UmiSchedule.constant_schedule(
            value=0, Name="AlwaysOff"
        )
        DataSource = kwargs.pop("DataSource", surface.theidf.name)
        Category = kwargs.pop("Category", surface.theidf.name)
        w = cls(
            Name=surface.Name,
            Construction=construction,
            idf=surface.theidf,
            Category=Category,
            DataSource=DataSource,
            **attr,
            **kwargs,
        )
        return w

    @classmethod
    @timeit
    def from_zone(cls, zone, **kwargs):
        """Iterate over the zone subsurfaces and create a window object.

        If more than one window is created, use reduce to combine them together.

        Args:
            zone (Zone): The Zone object from which the WindowSetting is
                created.

        Returns:
            WindowSetting: The WindowSetting object for this zone.
        """
        window_sets = []

        for surf in zone.zone_surfaces:
            # skip internalmass objects since they don't have windows.
            if surf.key.lower() != "internalmass":
                for subsurf in surf.subsurfaces:
                    # For each subsurface, create a WindowSetting object
                    # using the `from_surface` constructor.
                    window_sets.append(cls.from_surface(subsurf, **kwargs))

        if window_sets:
            # if one or more window has been created, reduce. Using reduce on
            # a len==1 list, will simply return the object.

            return reduce(WindowSetting.combine, window_sets)
        else:
            # no window found, probably a core zone, return None.
            return None

    def combine(self, other, weights=None, allow_duplicates=False):
        """Append other to self. Return self + other as a new object.

        Args:
            other (WindowSetting): The other OpaqueMaterial object
            weights (list-like, optional): A list-like object of len 2. If None,
                equal weights are used.

        Returns:
            WindowSetting: A new combined object made of self + other.
        """
        # Check if other is None. Simply return self
        if not other:
            return self

        if not self:
            return other

        if not isinstance(other, self.__class__):
            msg = "Cannot combine %s with %s" % (
                self.__class__.__name__,
                other.__class__.__name__,
            )
            raise NotImplementedError(msg)

        # Check if other is not the same as self
        if self == other:
            return self

        if not weights:
            log(
                'using 1 as weighting factor in "{}" '
                "combine.".format(self.__class__.__name__)
            )
            weights = [1.0, 1.0]
        meta = self._get_predecessors_meta(other)
        new_attr = dict(
            Construction=WindowConstruction.combine(
                self.Construction, other.Construction, weights
            ),
            AfnDischargeC=self.float_mean(other, "AfnDischargeC", weights),
            AfnTempSetpoint=self.float_mean(other, "AfnTempSetpoint", weights),
            AfnWindowAvailability=UmiSchedule.combine(
                self.AfnWindowAvailability, other.AfnWindowAvailability, weights
            ),
            IsShadingSystemOn=any([self.IsShadingSystemOn, other.IsShadingSystemOn]),
            IsVirtualPartition=any([self.IsVirtualPartition, other.IsVirtualPartition]),
            IsZoneMixingOn=any([self.IsZoneMixingOn, other.IsZoneMixingOn]),
            OperableArea=self.float_mean(other, "OperableArea", weights),
            ShadingSystemSetpoint=self.float_mean(
                other, "ShadingSystemSetpoint", weights
            ),
            ShadingSystemTransmittance=self.float_mean(
                other, "ShadingSystemTransmittance", weights
            ),
            ShadingSystemType=max(self.ShadingSystemType, other.ShadingSystemType),
            ZoneMixingDeltaTemperature=self.float_mean(
                other, "ZoneMixingDeltaTemperature", weights
            ),
            ZoneMixingFlowRate=self.float_mean(other, "ZoneMixingFlowRate", weights),
            ZoneMixingAvailabilitySchedule=UmiSchedule.combine(
                self.ZoneMixingAvailabilitySchedule,
                other.ZoneMixingAvailabilitySchedule,
                weights,
            ),
            ShadingSystemAvailabilitySchedule=UmiSchedule.combine(
                self.ShadingSystemAvailabilitySchedule,
                other.ShadingSystemAvailabilitySchedule,
                weights,
            ),
            Type=max(self.Type, other.Type),
        )
        new_obj = WindowSetting(**meta, **new_attr)
        new_obj.predecessors.update(self.predecessors + other.predecessors)
        return new_obj

    def to_dict(self):
        """Return WindowSetting dictionary representation."""
        self.validate()  # Validate object before trying to get json format

        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["AfnDischargeC"] = self.AfnDischargeC
        data_dict["AfnTempSetpoint"] = self.AfnTempSetpoint
        data_dict["AfnWindowAvailability"] = self.AfnWindowAvailability.to_ref()
        data_dict["Construction"] = {"$ref": str(self.Construction.id)}
        data_dict["IsShadingSystemOn"] = self.IsShadingSystemOn
        data_dict["IsVirtualPartition"] = self.IsVirtualPartition
        data_dict["IsZoneMixingOn"] = self.IsZoneMixingOn
        data_dict["OperableArea"] = self.OperableArea
        data_dict[
            "ShadingSystemAvailabilitySchedule"
        ] = self.ShadingSystemAvailabilitySchedule.to_ref()
        data_dict["ShadingSystemSetpoint"] = self.ShadingSystemSetpoint
        data_dict["ShadingSystemTransmittance"] = self.ShadingSystemTransmittance
        data_dict["ShadingSystemType"] = self.ShadingSystemType.value
        data_dict["Type"] = self.Type.value
        data_dict[
            "ZoneMixingAvailabilitySchedule"
        ] = self.ZoneMixingAvailabilitySchedule.to_ref()
        data_dict["ZoneMixingDeltaTemperature"] = self.ZoneMixingDeltaTemperature
        data_dict["ZoneMixingFlowRate"] = self.ZoneMixingFlowRate
        data_dict["Category"] = self.Category
        data_dict["Comments"] = validators.string(self.Comments, allow_empty=True)
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = self.Name

        return data_dict

    @classmethod
    def from_dict(cls, data, schedules, window_constructions, **kwargs):
        """Create a ZoneConditioning from a dictionary.

        Args:
            data (dict): The python dictionary.
            schedules (dict): A dictionary of UmiSchedules with their id as keys.
            window_constructions (dict): A dictionary of WindowConstruction objects
                with their id as keys.
            **kwargs: keywords passed to parent constructor.
        """
        data = copy(data)
        _id = data.pop("$id")
        afn_availability_schedule = schedules[data.pop("AfnWindowAvailability")["$ref"]]
        construction = window_constructions[data.pop("Construction")["$ref"]]
        shading_system_availability_schedule = schedules[
            data.pop("ShadingSystemAvailabilitySchedule")["$ref"]
        ]
        zone_mixing_availability_schedule = schedules[
            data.pop("ZoneMixingAvailabilitySchedule")["$ref"]
        ]
        return cls(
            id=_id,
            Construction=construction,
            AfnWindowAvailability=afn_availability_schedule,
            ShadingSystemAvailabilitySchedule=shading_system_availability_schedule,
            ZoneMixingAvailabilitySchedule=zone_mixing_availability_schedule,
            **data,
            **kwargs,
        )

    @classmethod
    def from_ref(
        cls, ref, building_templates, schedules, window_constructions, **kwargs
    ):
        """Initialize :class:`WindowSetting` object from a reference id.

        Hint:
            In some cases, the WindowSetting is referenced in the DataStore to the
            Windows property of a BuildingTemplate (instead of being listed in the
            WindowSettings list. This is the case in the original
            BostonTemplateLibrary.json.

        Args:
            ref (str): The referenced number in the json library.
            building_templates (list): List of BuildingTemplates from the datastore.

        Returns:
            WindowSetting: The parsed WindowSetting.
        """
        store = next(
            iter(
                filter(
                    lambda x: x.get("$id") == ref,
                    [bldg.get("Windows") for bldg in building_templates],
                )
            )
        )
        w = cls.from_dict(store, schedules, window_constructions, **kwargs)
        return w

    def validate(self):
        """Validate object and fill in missing values."""
        if self.AfnWindowAvailability is None:
            self.AfnWindowAvailability = UmiSchedule.constant_schedule(
                value=0, Name="AlwaysOff"
            )
        if self.ShadingSystemAvailabilitySchedule is None:
            self.ShadingSystemAvailabilitySchedule = UmiSchedule.constant_schedule(
                value=0, Name="AlwaysOff"
            )
        if self.ZoneMixingAvailabilitySchedule is None:
            self.ZoneMixingAvailabilitySchedule = UmiSchedule.constant_schedule(
                value=0, Name="AlwaysOff"
            )

        return self

    def mapping(self, validate=False):
        """Get a dict based on the object properties, useful for dict repr.

        Args:
            validate (bool): If True, try to validate object before returning the
                mapping.
        """
        if validate:
            self.validate()

        return dict(
            AfnDischargeC=self.AfnDischargeC,
            AfnTempSetpoint=self.AfnTempSetpoint,
            AfnWindowAvailability=self.AfnWindowAvailability,
            Construction=self.Construction,
            IsShadingSystemOn=self.IsShadingSystemOn,
            IsVirtualPartition=self.IsVirtualPartition,
            IsZoneMixingOn=self.IsZoneMixingOn,
            OperableArea=self.OperableArea,
            ShadingSystemAvailabilitySchedule=self.ShadingSystemAvailabilitySchedule,
            ShadingSystemSetpoint=self.ShadingSystemSetpoint,
            ShadingSystemTransmittance=self.ShadingSystemTransmittance,
            ShadingSystemType=self.ShadingSystemType,
            Type=self.Type,
            ZoneMixingAvailabilitySchedule=self.ZoneMixingAvailabilitySchedule,
            ZoneMixingDeltaTemperature=self.ZoneMixingDeltaTemperature,
            ZoneMixingFlowRate=self.ZoneMixingFlowRate,
            Category=self.Category,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    @property
    def children(self):
        return (
            self.AfnWindowAvailability,
            self.Construction,
            self.ShadingSystemAvailabilitySchedule,
            self.ZoneMixingAvailabilitySchedule,
        )



"""UmiSchedules module."""

import calendar
import collections
import hashlib
from datetime import datetime
from typing import List

import numpy as np
import pandas as pd
from validator_collection import validators

from archetypal.schedule import Schedule, _ScheduleParser, get_year_for_first_weekday
from archetypal.template.umi_base import UmiBase
from archetypal.utils import log


class UmiSchedule(Schedule, UmiBase):
    """Class that handles Schedules."""
    _CREATED_OBJECTS = []

    __slots__ = ("_quantity",)

    def __init__(self, Name, quantity=None, **kwargs):
        """Initialize object with parameters.

        Args:
            Name:
            quantity:
            **kwargs:
        """
        super(UmiSchedule, self).__init__(Name, **kwargs)
        self.quantity = quantity

        # Only at the end append self to _CREATED_OBJECTS
        self._CREATED_OBJECTS.append(self)

    @property
    def quantity(self):
        """Get or set the schedule quantity."""
        return self._quantity

    @quantity.setter
    def quantity(self, value):
        self._quantity = value

    @classmethod
    def constant_schedule(cls, value=1, Name="AlwaysOn", Type="Fraction", **kwargs):
        """Create an UmiSchedule with a constant value at each timestep.

        Args:
            Type:
            value (float):
            Name:
            idf:
            **kwargs:
        """
        value = validators.float(value)
        return super(UmiSchedule, cls).constant_schedule(
            value=value, Name=Name, Type=Type, **kwargs
        )

    @classmethod
    def random(cls, Name="AlwaysOn", Type="Fraction", **kwargs):
        """Create an UmiSchedule with a randomized value (0-1) at each timestep.

        Args:
            Name (str): The name of the Schedule.
            Type (str or ScheduleTypeLimits):
            **kwargs: keywords passed to the constructor.
        """
        values = np.random.rand(
            8760,
        )
        return cls(Values=values.tolist(), Name=Name, Type=Type, **kwargs)

    @classmethod
    def from_values(cls, Name, Values, Type="Fraction", **kwargs):
        """Create an UmiSchedule from a list of values.

        Args:
            Name (str): The name of the Schedule.
            Values (list):
            Type:
            **kwargs:
        """
        return super(UmiSchedule, cls).from_values(
            Name=Name, Values=Values, Type=Type, **kwargs
        )

    def combine(self, other, weights=None, quantity=None):
        """Combine two UmiSchedule objects together.

        Args:
            other (UmiSchedule): The other Schedule object to combine with.
            weights (list, dict or string): Attribute of self and other containing the
                weight factor. If a list is passed, it must have len = 2; the first
                element is applied to self and the second element is applied to other.
                If a dict is passed, the self.Name and other.Name are the keys. If a
                str is passed, the
            quantity (list or dict or bool): Scalar value that will be multiplied by
                self before the averaging occurs. This ensures that the resulting
                schedule returns the correct integrated value. If a dict is passed,
                keys are schedules Names and values are quantities.

        Returns:
            (UmiSchedule): the combined UmiSchedule object.

        Raises:
            TypeError: if Quantity is not of type list, tuple, dict or a callable.
        """
        # Check if other is None. Simply return self
        if not other:
            return self

        if not self:
            return other

        if not isinstance(other, UmiSchedule):
            msg = "Cannot combine %s with %s" % (
                self.__class__.__name__,
                other.__class__.__name__,
            )
            raise NotImplementedError(msg)

        # check if the schedule is the same
        if self == other:
            if self.quantity and other.quantity:
                self.quantity += other.quantity
            return self
        # check if self is only zeros. Should not affect other.
        if not np.any(self.all_values):
            return other
        # check if other is only zeros. Should not affect self.
        if not np.any(other.all_values):
            return self

        if not weights:
            log(
                'using 1 as weighting factor in "{}" '
                "combine.".format(self.__class__.__name__)
            )
            weights = [1, 1]
        elif isinstance(weights, str):
            # get the attribute from self and other
            weights = [getattr(self, weights), getattr(other, weights)]
        elif isinstance(weights, (list, tuple)):
            # check if length is 2.
            length = len(weights)
            if length != 2:
                raise ValueError(
                    "USing a list or tuple, the weights attribute must "
                    "have a length of 2. A length of {}".format(length)
                )
        elif isinstance(weights, dict):
            weights = [weights[self.Name], weights[other.Name]]

        if quantity is None:
            new_values = np.average(
                [self.all_values, other.all_values], axis=0, weights=weights
            )
        elif isinstance(quantity, dict):
            # Multiplying the schedule values by the quantity for both self and other
            # and then using a weighted average. Finally, new values are normalized.
            new_values = np.average(
                [
                    self.all_values * quantity[self.Name],
                    other.all_values * quantity[other.Name],
                ],
                axis=0,
                weights=weights,
            )
            new_values /= quantity[self.Name] + quantity[other.Name]
        elif callable(quantity):
            new_values = np.average(
                np.stack((self.all_values, other.all_values), axis=1),
                axis=1,
                weights=[
                    quantity(self.predecessors.data),
                    quantity(other.predecessors.data),
                ],
            )
        elif isinstance(quantity, (list, tuple)):
            # Multiplying the schedule values by the quantity for both self and other
            # and then using a weighted average. Finally, new values are normalized.
            self_quantity, other_quantity = quantity
            new_values = (
                self.all_values * self_quantity + other.all_values * other_quantity
            ) / sum(quantity)
        elif isinstance(quantity, bool):
            new_values = np.average(
                [self.all_values, other.all_values],
                axis=0,
                weights=[self.quantity * weights[0], other.quantity * weights[1]],
            )
        else:
            raise TypeError("Quantity is not of type list, tuple, dict or a callable")

        # the new object's name
        meta = self._get_predecessors_meta(other)

        # Overriding meta Name
        hasher = hashlib.md5()
        hasher.update(new_values)
        meta["Name"] = f"Combined_UmiSchedule_{hasher.hexdigest()}"
        quantity = np.nansum(
            [self.quantity or float("nan"), other.quantity or float("nan")]
        )
        new_obj = UmiSchedule.from_values(
            Values=new_values, Type="Fraction", quantity=quantity, **meta
        )
        new_obj.predecessors.update(self.predecessors + other.predecessors)
        new_obj.weights = sum(weights)
        return new_obj

    def develop(self):
        """Develop the UmiSchedule into a Year-Week-Day schedule structure."""
        year, weeks, days = self.to_year_week_day()
        lines = ["- {}".format(obj) for obj in self.predecessors]

        _from = "\n".join(lines)
        year.Comments = (
            f"Year Week Day schedules created from: \n{_from}" + str(id(self)),
        )
        year.quantity = self.quantity
        return year

    def get_unique(self):
        """Return the first of all the created objects that is equivalent to self."""
        return super(UmiSchedule, self.develop()).get_unique()

    def to_dict(self):
        """Return UmiSchedule dictionary representation.

        Hint:
            UmiSchedule does not implement the to_dict method because it is not used
            when generating the json file. Only Year-Week- and DaySchedule classes
            are used.
        """
        return self.to_ref()

    def to_ref(self):
        """Return a ref pointer to self."""
        return {"$ref": str(self.id)}

    def validate(self):
        """Validate object and fill in missing values."""
        return self

    def mapping(self, validate=False):
        """Get a dict based on the object properties, useful for dict repr.

        Args:
            validate (bool): If True, try to validate object before returning the
                mapping.
        """
        if validate:
            self.validate()

        return dict(
            Category=self.Category,
            Type=self.Type,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    def get_ref(self, ref):
        """Get item matching reference id.

        Args:
            ref:
        """
        return next(
            iter(
                [
                    value
                    for value in UmiSchedule.CREATED_OBJECTS
                    if value.id == ref["$ref"]
                ]
            ),
            None,
        )

    def duplicate(self):
        """Get copy of self."""
        return self.__copy__()

    def __add__(self, other):
        """Return new object that is the combination of self and other."""
        return UmiSchedule.combine(self, other)

    def __repr__(self):
        """Return a representation of self."""
        name = self.Name
        resample = self.series.resample("D")
        min = resample.min().mean()
        mean = resample.mean().mean()
        max = resample.max().mean()
        return (
            name
            + ": "
            + "mean daily min:{:.2f} mean:{:.2f} max:{:.2f} ".format(min, mean, max)
            + (f"quantity {self.quantity}" if self.quantity is not None else "")
        )

    def __str__(self):
        """Return the string representation of self."""
        return repr(self)

    def __hash__(self):
        """Return the hash value of self."""
        return hash(self.id)

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, UmiSchedule):
            return NotImplemented
        if self.all_values.size != other.all_values.size:
            return NotImplemented
        else:
            return all(
                [
                    self.strict == other.strict,
                    self.Type == other.Type,
                    self.quantity == other.quantity,
                    np.allclose(self.all_values, other.all_values, rtol=1e-02),
                ]
            )

    def __copy__(self):
        """Create a copy of self."""
        return self.__class__(
            Name=self.Name,
            quantity=self.quantity,
            Values=self.all_values.tolist(),
            strict=self.strict,
            Type=self.Type,
        )


class YearSchedulePart:
    """Helper Class for YearSchedules defined with FromDay FromMonth ToDay ToMonth."""

    __slots__ = ("_from_day", "_from_month", "_to_day", "_to_month", "_schedule")

    def __init__(
        self,
        FromDay=None,
        FromMonth=None,
        ToDay=None,
        ToMonth=None,
        Schedule=None,
        **kwargs,
    ):
        """Initialize YearSchedulePart.

        Args:
            FromDay (int): This numeric field is the starting day for the
                schedule time period.
            FromMonth (int): This numeric field is the starting month for the
                schedule time period.
            ToDay (int): This numeric field is the ending day for the schedule
                time period.
            ToMonth (int): This numeric field is the ending month for the
                schedule time period.
            Schedule (UmiSchedule): The associated UmiSchedule related to this
                object.
            kwargs (dict): Other Keyword arguments.
        """
        self.FromDay = FromDay
        self.FromMonth = FromMonth
        self.ToDay = ToDay
        self.ToMonth = ToMonth
        self.Schedule = Schedule

    @property
    def FromDay(self):
        """Get or set the start day-of-month number [int]."""
        return self._from_day

    @FromDay.setter
    def FromDay(self, value):
        self._from_day = validators.integer(value, minimum=1, maximum=31)

    @property
    def FromMonth(self):
        """Get or set the start month-number [int]."""
        return self._from_month

    @FromMonth.setter
    def FromMonth(self, value):
        self._from_month = validators.integer(value, minimum=1, maximum=12)

    @property
    def ToDay(self):
        """Get or set the end day-of-month number [int]."""
        return self._to_day

    @ToDay.setter
    def ToDay(self, value):
        self._to_day = validators.integer(value, minimum=1, maximum=31)

    @property
    def ToMonth(self):
        """Get or set the end month-number [int]."""
        return self._to_month

    @ToMonth.setter
    def ToMonth(self, value):
        self._to_month = validators.integer(value, minimum=1, maximum=12)

    @property
    def Schedule(self):
        """Get or set the WeekSchedule object."""
        return self._schedule

    @Schedule.setter
    def Schedule(self, value):
        assert isinstance(value, WeekSchedule), "schedule must be of type WeekSchedule"
        self._schedule = value

    @classmethod
    def from_dict(cls, data, schedules, **kwargs):
        """Create a YearSchedulePart object from a dictionary.

        Args:
            data (dict): The python dictionary.
            schedules (dict): A dictionary of WeekSchedules with their id as keys.
            **kwargs: keywords passed to parent constructor.

        .. code-block:: python

            data = {
                'FromDay': 1,
                'FromMonth': 1,
                'ToDay': 31,
                'ToMonth': 12,
                'Schedule': {'$ref': '140622440042800'}
            }
        """
        schedule = schedules[data.pop("Schedule")["$ref"]]
        ysp = cls(Schedule=schedule, **data, **kwargs)

        return ysp

    def to_dict(self):
        """Return YearSchedulePart dictionary representation."""
        return collections.OrderedDict(
            FromDay=self.FromDay,
            FromMonth=self.FromMonth,
            ToDay=self.ToDay,
            ToMonth=self.ToMonth,
            Schedule={"$ref": str(self.Schedule.id)},
        )

    def __str__(self):
        """Return string representation of self."""
        return repr(self)

    def __repr__(self):
        return "".join([f"{k}={v}" for k, v in self.to_dict().items()])

    def mapping(self):
        """Get a dict based on the object properties, useful for dict repr."""
        return dict(
            FromDay=self.FromDay,
            FromMonth=self.FromMonth,
            ToDay=self.ToDay,
            ToMonth=self.ToMonth,
            Schedule=self.Schedule,
        )

    def get_unique(self):
        """Return the first of all the created objects that is equivalent to self."""
        return self

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, YearSchedulePart):
            return NotImplemented
        else:
            return all(
                [
                    self.FromDay == other.FromDay,
                    self.FromMonth == other.FromMonth,
                    self.ToDay == other.ToDay,
                    self.ToMonth == other.ToMonth,
                    self.Schedule == other.Schedule,
                ]
            )

    def __iter__(self):
        """Iterate over attributes. Yields tuple of (keys, value)."""
        for k, v in self.mapping().items():
            yield k, v

    def __hash__(self):
        """Return the hash value of self."""
        return id(self)


class DaySchedule(UmiSchedule):
    """Superclass of UmiSchedule that handles daily schedules."""

    __slots__ = ("_values",)

    def __init__(self, Name, Values, Category="Day", **kwargs):
        """Initialize a DaySchedule object with parameters.

        Args:
            Values (list): List of 24 values.
            Name (str): Name of the schedule.
            Category (str): category identification (default: "Day").
            **kwargs: Keywords passed to the :class:`UmiSchedule` constructor.
        """
        super(DaySchedule, self).__init__(
            Category=Category, Name=Name, Values=Values, **kwargs
        )

    @property
    def all_values(self) -> np.ndarray:
        """Return numpy array of schedule Values."""
        return np.array(self._values)

    @all_values.setter
    def all_values(self, value):
        self._values = validators.iterable(value, maximum_length=24)

    @classmethod
    def from_epbunch(cls, epbunch, strict=False, **kwargs):
        """Create a DaySchedule from an EpBunch.

        This method accepts "Schedule:Day:Hourly", "Schedule:Day:List" and
        "Schedule:Day:Interval".

        Args:
            epbunch (EpBunch): The EpBunch object to construct a DaySchedule
                from.
            **kwargs: Keywords passed to the :class:`UmiSchedule` constructor.
                See :class:`UmiSchedule` for more details.
        """
        assert epbunch.key.lower() in (
            "schedule:day:hourly",
            "schedule:day:list",
            "schedule:day:interval",
        ), (
            f"Input error for '{epbunch.key}'. Expected on of "
            f"('Schedule:Day:Hourly', 'Schedule:Day:List' and "
            f"'Schedule:Day:Interval')"
        )
        start_day_of_the_week = epbunch.theidf.day_of_week_for_start_day
        start_date = datetime(get_year_for_first_weekday(start_day_of_the_week), 1, 1)
        sched = cls(
            Name=epbunch.Name,
            epbunch=epbunch,
            schType=epbunch.key,
            Type=cls.get_schedule_type_limits_name(epbunch),
            Values=_ScheduleParser.get_schedule_values(
                epbunch, start_date, strict=strict
            ),
            **kwargs,
        )

        return sched

    @classmethod
    def from_values(cls, Name, Values, Type="Fraction", **kwargs):
        """Create a DaySchedule from an array of size (24,).

        Args:
            Name:
            Values (array-like): A list of values of length 24.
            Type (str): Schedule Type Limit name.
            **kwargs: Keywords passed to the :class:`UmiSchedule` constructor.
                See :class:`UmiSchedule` for more details.
        """
        return cls(Name=Name, Values=Values, Type=Type, **kwargs)

    @classmethod
    def from_dict(cls, data, **kwargs):
        """Create a DaySchedule from a dictionary.

        Args:
            data (dict): A python dictionary with the structure shown bellow.
            **kwargs: keywords passed to parents constructors.

        .. code-block:: python

            {
              "$id": "67",
              "Category": "Day",
              "Type": "Fraction",
              "Values": [...],  # 24 hourly values
              "Comments": "default",
              "DataSource": "default",
              "Name": "B_Res_D_Occ_WD"
            },
        """
        _id = data.pop("$id")
        sched = cls.from_values(id=_id, **data, **kwargs)

        return sched

    def get_unique(self):
        """Return the first of all the created objects that is equivalent to self."""
        return UmiBase.get_unique(self)

    def to_dict(self):
        """Return DaySchedule dictionary representation."""
        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["Category"] = self.Category
        data_dict["Type"] = "Fraction" if self.Type is None else self.Type.Name
        data_dict["Values"] = np.round(self.all_values, 3).tolist()
        data_dict["Comments"] = validators.string(self.Comments, allow_empty=True)
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = self.Name

        return data_dict

    def mapping(self, validate=False):
        """Get a dict based on the object properties, useful for dict repr.

        Args:
            validate (bool): If True, try to validate object before returning the
                mapping.
        """
        if validate:
            self.validate()

        return dict(
            Category=self.Category,
            Type=self.Type,
            Values=self.all_values.round(3).tolist(),
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    def to_ref(self):
        """Return a ref pointer to self."""
        return {"$ref": str(self.id)}

    def duplicate(self):
        """Get copy of self."""
        return self.__copy__()

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, DaySchedule):
            return NotImplemented
        else:
            return all(
                [
                    self.Type == other.Type,
                    np.allclose(self.all_values, other.all_values, rtol=1e-02),
                ]
            )

    def __hash__(self):
        """Return the hash value of self."""
        return super(DaySchedule, self).__hash__()

    def __copy__(self):
        """Create a copy of self."""
        return self.__class__(self.Name, Values=self.all_values.tolist())

    def to_epbunch(self, idf):
        """Convert self to an epbunch given an idf model.

        Args:
            idf (IDF): An IDF model.

        .. code-block:: python

            SCHEDULE:DAY:HOURLY,
                ,                         !- Name
                ,                         !- Schedule Type Limits Name
                0,                        !- Hour 1
                0,                        !- Hour 2
                0,                        !- Hour 3
                0,                        !- Hour 4
                0,                        !- Hour 5
                0,                        !- Hour 6
                0,                        !- Hour 7
                0,                        !- Hour 8
                0,                        !- Hour 9
                0,                        !- Hour 10
                0,                        !- Hour 11
                0,                        !- Hour 12
                0,                        !- Hour 13
                0,                        !- Hour 14
                0,                        !- Hour 15
                0,                        !- Hour 16
                0,                        !- Hour 17
                0,                        !- Hour 18
                0,                        !- Hour 19
                0,                        !- Hour 20
                0,                        !- Hour 21
                0,                        !- Hour 22
                0,                        !- Hour 23
                0;                        !- Hour 24

        Returns:
            EpBunch: The EpBunch object added to the idf model.
        """
        return idf.newidfobject(
            key="Schedule:Day:Hourly".upper(),
            **dict(
                Name=self.Name,
                Schedule_Type_Limits_Name=self.Type.to_epbunch(idf).Name,
                **{"Hour_{}".format(i + 1): self.all_values[i] for i in range(24)},
            ),
        )


class WeekSchedule(UmiSchedule):
    """Superclass of UmiSchedule that handles weekly schedules."""

    __slots__ = ("_days", "_values")

    def __init__(self, Name, Days=None, Category="Week", **kwargs):
        """Initialize a WeekSchedule object with parameters.

        Args:
            Days (list of DaySchedule): list of :class:`DaySchedule`.
            **kwargs:
        """
        super(WeekSchedule, self).__init__(Name, Category=Category, **kwargs)
        self.Days = Days

    @property
    def Days(self):
        """Get or set the list of DaySchedule objects."""
        return self._days

    @Days.setter
    def Days(self, value):
        if value is not None:
            assert all(
                isinstance(x, DaySchedule) for x in value
            ), f"Input value error '{value}'. Expected list of DaySchedule."
        self._days = value

    @classmethod
    def from_epbunch(cls, epbunch, **kwargs):
        """Create a WeekSchedule from a Schedule:Week:Daily object.

        Args:
            epbunch (EpBunch): The Schedule:Week:Daily object.
            **kwargs: keywords passed to the constructor.
        """
        assert (
            epbunch.key.lower() == "schedule:week:daily"
        ), f"Expected a 'schedule:week:daily' not a '{epbunch.key.lower()}'"
        Days = WeekSchedule.get_days(epbunch, **kwargs)
        sched = cls(
            Name=epbunch.Name,
            schType=epbunch.key,
            Days=Days,
            **kwargs,
        )

        return sched

    @classmethod
    def from_dict(cls, data, day_schedules, **kwargs):
        """Create a WeekSchedule from a dictionary.

        Args:
            data (dict): The python dictionary.
            day_schedules (dict): A dictionary of python DaySchedules with their id as
                keys.
            **kwargs: keywords passed to the constructor.
        """
        refs = data.pop("Days")
        _id = data.pop("$id")
        Days = [day_schedules[ref["$ref"]] for ref in refs]
        wc = cls(Days=Days, id=_id, **data, **kwargs)
        return wc

    def get_unique(self):
        """Return the first of all the created objects that is equivalent to self."""
        return UmiBase.get_unique(self)

    def to_dict(self):
        """Return WeekSchedule dictionary representation."""
        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["Category"] = self.Category
        data_dict["Days"] = [day.to_ref() for day in self.Days]
        data_dict["Type"] = "Fraction" if self.Type is None else self.Type.Name
        data_dict["Comments"] = validators.string(self.Comments, allow_empty=True)
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = self.Name

        return data_dict

    def mapping(self, validate=False):
        """Get a dict based on the object properties, useful for dict repr.

        Args:
            validate (bool): If True, try to validate object before returning the
                mapping.
        """
        if validate:
            self.validate()

        return dict(
            Category=self.Category,
            Days=self.Days,
            Type=self.Type,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    @classmethod
    def get_days(cls, epbunch, **kwargs):
        """Get the DaySchedules referenced in the Week:Schedule:Days object.

        Args:
            list of DaySchedule: The list of DaySchedules referenced by the epbunch.
        """
        assert (
            epbunch.key.lower() == "schedule:week:daily"
        ), f"Expected a 'schedule:week:daily' not a '{epbunch.key.lower()}'"
        Days = []
        dayname = [
            "Monday",
            "Tuesday",
            "Wednesday",
            "Thursday",
            "Friday",
            "Saturday",
            "Sunday",
        ]
        for day in dayname:
            day_ep = epbunch.get_referenced_object("{}_ScheduleDay_Name".format(day))
            Days.append(DaySchedule.from_epbunch(day_ep, **kwargs))

        return Days

    @property
    def all_values(self) -> np.ndarray:
        """Return numpy array of schedule Values."""
        if self._values is None:
            self._values = np.concatenate([day.all_values for day in self.Days])
        return self._values

    def to_ref(self):
        """Return a ref pointer to self."""
        return {"$ref": str(self.id)}

    def duplicate(self):
        """Get copy of self."""
        return self.__copy__()

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, WeekSchedule):
            return NotImplemented
        else:
            return all(
                [
                    self.Type == other.Type,
                    self.Days == other.Days,
                ]
            )

    def __hash__(self):
        """Return the hash value of self."""
        return super(WeekSchedule, self).__hash__()

    def __copy__(self):
        """Create a copy of self."""
        return self.__class__(Name=self.Name, Days=self.Days)

    def to_epbunch(self, idf):
        """Convert self to an epbunch given an idf model.

        Args:
            idf (IDF): An IDF model.

        Returns:
            EpBunch: The EpBunch object added to the idf model.
        """
        return idf.newidfobject(
            key="Schedule:Week:Daily".upper(),
            **dict(
                Name=self.Name,
                **{
                    f"{calendar.day_name[i]}_ScheduleDay_Name": day.to_epbunch(idf).Name
                    for i, day in enumerate(self.Days)
                },
                Holiday_ScheduleDay_Name=self.Days[6].Name,
                SummerDesignDay_ScheduleDay_Name=self.Days[0].Name,
                WinterDesignDay_ScheduleDay_Name=self.Days[0].Name,
                CustomDay1_ScheduleDay_Name=self.Days[1].Name,
                CustomDay2_ScheduleDay_Name=self.Days[6].Name,
            ),
        )

    @property
    def children(self):
        return self.Days


class YearSchedule(UmiSchedule):
    """Superclass of UmiSchedule that handles yearly schedules."""

    def __init__(self, Name, Type="Fraction", Parts=None, Category="Year", **kwargs):
        """Initialize a YearSchedule object with parameters.

        Args:
            Category:
            Name:
            Type:
            Parts (list of YearSchedulePart): The YearScheduleParts.
            **kwargs:
        """
        self.epbunch = kwargs.get("epbunch", None)
        if Parts is None:
            self.Parts = self._get_parts(self.epbunch)
        else:
            self.Parts = Parts
        super(YearSchedule, self).__init__(
            Name=Name, Type=Type, schType="Schedule:Year", Category=Category, **kwargs
        )

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, YearSchedule):
            return NotImplemented
        else:
            return all([self.Type == other.Type, self.Parts == other.Parts])

    def __hash__(self):
        """Return the hash value of self."""
        return super(YearSchedule, self).__hash__()

    @property
    def all_values(self) -> np.ndarray:
        """Return numpy array of schedule Values."""
        if self._values is None:
            index = pd.date_range(start=self.startDate, freq="1H", periods=8760)
            series = pd.Series(index=index)
            for part in self.Parts:
                start = "{}-{}-{}".format(self.year, part.FromMonth, part.FromDay)
                end = "{}-{}-{}".format(self.year, part.ToMonth, part.ToDay)
                # Get week values from all_values of Days
                one_week = np.array(
                    [
                        item
                        for sublist in part.Schedule.Days
                        for item in sublist.all_values
                    ]
                )

                all_weeks = np.resize(one_week, len(series.loc[start:end]))
                series.loc[start:end] = all_weeks
            self._values = series.values
        return self._values

    @classmethod
    def from_dict(cls, data, week_schedules, **kwargs):
        """Create a YearSchedule from a dictionary.

        Args:
            data (dict): The python dictionary.
            week_schedule (dict): A dictionary of python WeekSchedules with their id as
                keys.
            **kwargs: keywords passed to the constructor.
        """
        Parts: List[YearSchedulePart] = [
            YearSchedulePart.from_dict(data, week_schedules)
            for data in data.pop("Parts", None)
        ]
        _id = data.pop("$id")
        ys = cls(Parts=Parts, id=_id, **data, **kwargs)
        return ys

    def get_unique(self):
        """Return the first of all the created objects that is equivalent to self."""
        return UmiBase.get_unique(self)

    def to_dict(self):
        """Return YearSchedule dictionary representation."""
        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["Category"] = self.Category
        data_dict["Parts"] = [part.to_dict() for part in self.Parts]
        data_dict["Type"] = "Fraction" if self.Type is None else self.Type.Name
        data_dict["Comments"] = validators.string(self.Comments, allow_empty=True)
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = self.Name

        return data_dict

    def to_epbunch(self, idf):
        """Convert self to an epbunch given an idf model.

        Notes:
            The object is added to the idf model.

        Args:
            idf (IDF): An IDF model.

        Returns:
            EpBunch: The EpBunch object added to the idf model.
        """
        new_dict = dict(
            Name=self.Name, Schedule_Type_Limits_Name=self.Type.to_epbunch(idf).Name
        )
        for i, part in enumerate(self.Parts):
            new_dict.update(
                {
                    "ScheduleWeek_Name_{}".format(i + 1): part.Schedule.to_epbunch(
                        idf
                    ).Name,
                    "Start_Month_{}".format(i + 1): part.FromMonth,
                    "Start_Day_{}".format(i + 1): part.FromDay,
                    "End_Month_{}".format(i + 1): part.ToMonth,
                    "End_Day_{}".format(i + 1): part.ToDay,
                }
            )

        return idf.newidfobject(key="Schedule:Year".upper(), **new_dict)

    def mapping(self, validate=False):
        """Get a dict based on the object properties, useful for dict repr.

        Args:
            validate (bool): If True, try to validate object before returning the
                mapping.
        """
        if validate:
            self.validate()

        return dict(
            Category=self.Category,
            Parts=self.Parts,
            Type=self.Type,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    def _get_parts(self, epbunch):
        parts = []
        for i in range(int(len(epbunch.fieldvalues[3:]) / 5)):
            week_day_schedule_name = epbunch["ScheduleWeek_Name_{}".format(i + 1)]

            FromMonth = epbunch["Start_Month_{}".format(i + 1)]
            ToMonth = epbunch["End_Month_{}".format(i + 1)]
            FromDay = epbunch["Start_Day_{}".format(i + 1)]
            ToDay = epbunch["End_Day_{}".format(i + 1)]
            parts.append(
                YearSchedulePart(
                    FromDay,
                    FromMonth,
                    ToDay,
                    ToMonth,
                    next(
                        (
                            x
                            for x in self._CREATED_OBJECTS
                            if x.Name == week_day_schedule_name
                            and type(x).__name__ == "WeekSchedule"
                        )
                    ),
                )
            )
        return parts

    def to_ref(self):
        """Return a ref pointer to self."""
        return {"$ref": str(self.id)}

    @property
    def children(self):
        return tuple(p.Schedule for p in self.Parts)
